const express = require("express");
const app = express();
const bodyParser = require("body-parser");
require("date-format-lite");
const { Client } = require("pg")


const port = 8080;
const db_options = {
    user: "postgres",
    host: "localhost",
    database: "gantt",
    password: "123",
    port: 5432,
}



const db = new Client(db_options)
db.connect(async (err) => {
    if (err) {
        console.error('connection error', err.stack);
    } else {
        console.log('connected');

        await db.query(`
            CREATE TABLE IF NOT EXISTS "gantt_links" (
              "id" SERIAL PRIMARY KEY,
              "source" integer NOT NULL,
              "target" integer NOT NULL,
              "type" varchar(1)
            );
        `)
        await db.query(`
            CREATE TABLE IF NOT EXISTS "gantt_tasks" (
              "id"SERIAL PRIMARY KEY,
              "text" varchar(255) NOT NULL,
              "start_date" TIMESTAMP NOT NULL,
              "duration" integer NOT NULL,
              "progress" float NOT NULL DEFAULT 0,
              "parent" integer NOT NULL,
              "sortorder" integer NOT NULL,
              "type" varchar(100),
              "color" varchar(100) DEFAULT ''
            );
        `)
    }
})


app.use(express.static("public"));
app.use(bodyParser.urlencoded({ extended: true }));

app.listen(port, function() {
    console.log("Server is running on port " + port);
});


// list task and links
app.get("/data", async (req, res) => {
    try {
        let { rows: tasks } = await db.query("SELECT * FROM gantt_tasks ORDER BY sortorder ASC")
        let { rows: links } = await db.query("SELECT * FROM gantt_links")

        tasks.forEach((task) => {
            task.start_date = task.start_date.format("YYYY-MM-DD hh:mm:ss");
            task.open = true;
        })

        res.send({
            data: tasks,
            collections: { links: links }
        });
    } catch (error) {
        sendResponse(res, "error", null, error);
    }
});


// add new task
app.post("/data/task", async (req, res) => { // adds new task to database
    const task = getTask(req.body);

    try {
        const { rows } = await db.query("SELECT MAX(sortorder) AS maxorder FROM gantt_tasks")
        const orderIndex = (rows[0].maxorder || 0) + 1;

        const result = await db.query("INSERT INTO gantt_tasks(text, start_date, duration, progress, parent, sortorder, type, color) VALUES ($1,$2,$3,$4,$5,$6,$7,$8) returning id",
            [task.text, task.start_date, task.duration, task.progress, task.parent, orderIndex, task.type, task.color]);
        const taskId = result.rows[0].id;
        sendResponse(res, "inserted", taskId);

    } catch (error) {
        sendResponse(res, "error", null, error);
    }
});

// update task
app.put("/data/task/:id", async (req, res) => {
    const sid = req.params.id,
        target = req.body.target,
        task = getTask(req.body);

    try {
        await db.query("UPDATE gantt_tasks SET text = $1, start_date = $2, duration = $3, progress = $4, parent = $5, type = $6, color = $7 WHERE id = $8",
            [task.text, task.start_date, task.duration, task.progress, task.parent, task.type, task.color, sid]);
        await updateOrder(sid, target)
        sendResponse(res, "updated");
    } catch (error) {
        sendResponse(res, "error", null, error);
    }
});

async function updateOrder(taskId, target) {
    let nextTask = false,
        targetOrder;

    target = target || "";

    if (target.startsWith("next:")) {
        target = target.substr("next:".length);
        if (target == "null") target = undefined;
        nextTask = true;
    }

    if (!target) return;

    const { rows } = await db.query("SELECT * FROM gantt_tasks WHERE id = $1", [target])
    if (!rows[0]) return;

    targetOrder = rows[0].sortorder;
    if (nextTask) targetOrder++;

    await db.query("UPDATE gantt_tasks SET sortorder = sortorder + 1 WHERE sortorder >= $1", [targetOrder])
    await db.query("UPDATE gantt_tasks SET sortorder = $1 WHERE id = $2", [targetOrder, taskId]);
}

// delete task
app.delete("/data/task/:id", async (req, res) => {
    const sid = req.params.id;
    try {
        await db.query("DELETE FROM gantt_tasks WHERE id = $1", [sid]);
        sendResponse(res, "deleted");
    } catch (error) {
        sendResponse(res, "error", null, error);
    }
});

// add link
app.post("/data/link", async (req, res) => {
    const link = getLink(req.body);

    try {
        const result = await db.query("INSERT INTO gantt_links(source, target, type) VALUES ($1,$2,$3) returning id",
            [link.source, link.target, link.type]);
        const linkId = result.rows[0].id;
        sendResponse(res, "inserted", linkId);
    } catch (error) {
        sendResponse(res, "error", null, error);
    }
});

// update link
app.put("/data/link/:id", async (req, res) => {
    const sid = req.params.id,
        link = getLink(req.body);

    try {
        await db.query("UPDATE gantt_links SET source = $1, target = $2, type = $3 WHERE id = $4",
            [link.source, link.target, link.type, sid])
        sendResponse(res, "updated");
    } catch (error) {
        sendResponse(res, "error", null, error);
    }
});

// delete link
app.delete("/data/link/:id", async (req, res) => {
    const sid = req.params.id;

    try {
        await db.query("DELETE FROM gantt_links WHERE id = $1", [sid])
        sendResponse(res, "deleted");
    } catch (error) {
        sendResponse(res, "error", null, error);
    }
});


function getTask(data) {
    return {
        text: data.text,
        start_date: data.start_date.date("YYYY-MM-DD"),
        duration: data.duration,
        progress: data.progress || 0,
        parent: data.parent,
        type: data.type,
        color: data.color
    };
}

function getLink(data) {
    return {
        source: data.source,
        target: data.target,
        type: data.type
    };
}

function sendResponse(res, action, tid, error) {
    if (action == "error")
        console.log(error);

    var result = {
        action: action
    };
    if (tid !== undefined && tid !== null)
        result.tid = tid;

    res.send(result);
}